import re
from urllib.parse import urlparse
from tld import get_fld

class BluesURL():

  @classmethod
  def get_domain(self,url):
    parsed_url = urlparse(url)
    return parsed_url.netloc

  @classmethod
  def get_main_domain(self,url):
    return get_fld(url)


class BluesType():
  @classmethod
  def last_index(cls,sequence,value):
    '''
    @description : get the last index of a value in a sequence
    @param {tuple|list} sequence
    @param {any} value
    @returns {init} : Returns -1 if there is no match
    '''
    try:
      return (len(sequence) - 1) - sequence[::-1].index(value)
    except ValueError:
      return -1 
 


import os,sys,re,logging,datetime
from .BluesFiler import BluesFiler 
from .BluesMailer import BluesMailer 

class BluesLogger():

  LOG_LEVELS = {
    'debug':logging.DEBUG,
    'info':logging.INFO, # will print all system info 
    'warning':logging.WARNING,
    'error':logging.ERROR,
  }

  DEFAULT_CONFIG = {
    'name':'system',
    'directory':'./log',
    'level':'info',
    'retention_days':7,
  }

  def __init__(self,config={}):
    '''
    @description : set log config
    @param {str} config.directory : Directory for storing logs
    @param {str} config.name : log's topic
    @param {int} config.retention_days : Maximum number of days for storing logs
    @param {str} config.level : The lowest level of log output
      - enum: debug info warning error
    '''
    self.config = {**self.DEFAULT_CONFIG,**config}
    self.log_level  = self.LOG_LEVELS[self.config['level']]
    self.logger = self.__get_logger()
    # clear log history
    BluesFiler.removefiles(self.config['directory'],self.config['retention_days'])

  def __get_logger(self):

    logging.basicConfig(level=self.log_level)

    logger = logging.getLogger(self.config['name'])

    log_file = self.__get_log_file()
    formatter = self.__get_formatter()
    file_logger = logging.FileHandler(log_file,'a','utf-8')
    file_logger.setFormatter(formatter)

    logger.addHandler(file_logger)

    return logger

  def __get_formatter(self):
    split_line = '\n\n'+''.join(['-' for x in range(70)])
    formatter = split_line+'\n\n'+'%(levelname)s (%(name)s) %(asctime)s:\n%(filename)s (line:%(lineno)s) %(funcName)s:\n%(message)s'
    return logging.Formatter(formatter)
      
  def __get_log_file(self):
    today = datetime.datetime.now().strftime("%Y-%m-%d")
    log_name = '%s.log' % (today)
    log_file = '%s\\%s' % (self.config['directory'],log_name)
    # make sure the dir exist
    BluesFiler.makedirs(self.config['directory'])
    return log_file

  def write(self,payload):
    '''
    @description write log
    @param {LogPayload} payload
     - message : str
     - level : str
     - mail : MailPayload
    '''

    if not payload.get('message'):
      return 

    level = payload.get('level','info')
    log_method = getattr(self.logger,level) if hasattr(self.logger,level) else self.logger.info
    log_method(payload.get('message'))

    if payload.get('mail') and payload.get('mail').get('addressee'):
      mail_payload = self.__get_mail_payload(payload)
      BluesMailer.send(mail_payload)

  def __get_mail_payload(self,payload):
    return {
      'subject':'%s %s : %s' % (self.config['name'],payload.get('level'),payload.get('mail').get('subject')),
      'content':payload.get('message'),
      'addressee':payload.get('mail').get('addressee'), 
      'addressee_name':payload.get('mail').get('addressee_name'),
    }



from .browser.BluesChrome import BluesChrome 
from .browser.BluesDebugChrome import BluesDebugChrome
from .browser.BluesProxyChrome import BluesProxyChrome
from .browser.BluesCookieChrome import BluesCookieChrome 

class BluesBrowser():

  @classmethod
  def chrome(cls,config={},arguments={},experimental_options={}):
    return BluesChrome(config,arguments,experimental_options)
  
  @classmethod
  def debug_chrome(cls,config={},arguments={},experimental_options={}):
    return BluesDebugChrome(config,arguments,experimental_options)
  
  @classmethod
  def proxy_chrome(cls,config={},arguments={},experimental_options={}):
    return BluesProxyChrome(config,arguments,experimental_options)

  @classmethod
  def cookie_chrome(cls,config={},arguments={},experimental_options={}):
    return BluesCookieChrome(config,arguments,experimental_options)


import time
from .BluesElementAction import BluesElementAction

# 提供窗口相关功能
class BluesDocumentAction():
 
  def __init__(self,driver):
    self.driver = driver
    self.element = BluesElementAction(driver)
  
  def shot(self,selector,file=None):
    '''
    @description 指定元素截图,不支持base64格式
    @param {str} selector : css selector 
    @param {str} file 保存位置
    @returns {str} file_path
    '''
    file_path = file if file else self.__get_default_file('elementshot')
    locator = self.element.wait(selector)
    shot_status = locator.screenshot(file_path)
    return file_path if shot_status else ''

  def __get_default_file(self,prefix='elementshot'):
    timestamp = int(time.time()) 
    return './%s-%s.png' % (prefix,timestamp)

  def get_attr(self,key):
    '''
    @descripton : get current document's info
    @param {str} key 
      - current_url
      - title
      - name
      - page_source
    '''
    return getattr(self.driver,key,'')


  


import sys,os,re
sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesURL import BluesURL  
from config.BluesConfig import BluesConfig   

# 提供Cookie相关功能
class BluesCookieAction():
 
  def __init__(self,driver):
    self.driver = driver

  def get(self,name=''):
    '''
    @description : get one cookie
    @param {str} name : cookie name
    @returns {dict} 形如：{'domain': 'mp.163.com', 'httpOnly': True, 'name': 'NTESwebSI', 'path': '/', 'sameSite': 'Lax', 'secure': False, 'value': '2A7C7F8FCD65F7D74650E13D349C60CC'}
    '''
    return self.driver.get_cookie(name)

  def get_all(self):
    '''
    @description : get all cookies
    @returns {dict[]} 形如：[{'domain': 'mp.163.com', 'httpOnly': True, 'name': 'NTESwebSI', 'path': '/', 'sameSite': 'Lax', 'secure': False, 'value': '2A7C7F8FCD65F7D74650E13D349C60CC'}]
    '''
    return self.driver.get_cookies()
      
  '''
  @description 添加cookie
  @param {dict} cookie 字典形式的cookie对象，至少包含(其他元素自动填充默认)：
    - name
    - value
  '''
  def set(self,cookie):
    if cookie.get('name') and ('value' in cookie):
      self.driver.add_cookie(cookie)

  '''
  @description 删除cookie
  @param {string} name cookie name
  '''
  def remove(self,name=''):
    return self.driver.delete_cookie(name)
  
  def clear(self):
    return self.driver.delete_all_cookies()

  def add_cookies(self,cookies):
    '''
    @description : add cookie from string or dict
    @param {str|dict}: 
     - 'BIDUPSID=77884ECAEE62BDD2A4A723BEF544DCB2; PSTM=1715957318'
     - {'BIDUPSID':'77884ECAEE62BDD2A4A723BEF544DCB2', 'PSTM':'1715957318'}
    @returns {None}
    '''
    if type(cookies) == str:
      cookie_dict = self.get_cookie_dict(cookies)
    else:
      cookie_dict = cookies

    if not cookie_dict:
      return 
    
    main_domain = BluesURL.get_main_domain(self.driver.current_url)
    for key,value in cookie_dict.items():
      self.set({
        'name':key,
        'value':str(value), # value's type must be string
        'domain':'.%s' % main_domain, # set main domain
      })

  def get_cookie_dict(self,cookies):
    '''
    @description : convert cookie str to dict
    @param {str}: 
     - 'BIDUPSID=77884ECAEE62BDD2A4A723BEF544DCB2; PSTM=1715957318'
    @returns {dict}
    '''
    if not cookies:
      return None

    cookie_dict = {}
    cookie_list = cookies.split(';')
    for k_v in cookie_list:
      kv=k_v.split('=')
      key = kv[0].strip()
      value = kv[1].strip()
      cookie_dict[key]=value
    return cookie_dict



import smtplib,datetime
from email.message import EmailMessage
from email.header import Header

class __BluesMailer():

  config = {
    # pick a smtp server
    'server' : 'smtp.qq.com',
    'port' : 465,
    # debug level: 0 - no message; 1 - many messages
    'debug_level' : 0, 
    # the sender's address
    'addresser' : '1121557826@qq.com',
    'addresser_name' : 'BluesLiu QQ',
    # the qq's auth code (not the account's login password)
    'addresser_pwd' : 'ryqokljshrrlifae',
  }

  def __init__(self):
    self.connection = self.__get_connection()

  def __get_connection(self):
    connection = smtplib.SMTP_SSL(self.config['server'],self.config['port'])
    connection.set_debuglevel(self.config['debug_level'])
    connection.login(self.config['addresser'],self.config['addresser_pwd'])
    return connection

  def send(self,payload):
    '''
    @description : send the mail
    @param {MailPayload} payload : mail's required info
     - addressee ：list | str  ; required
     - addressee_name ：str , can't contains space
     - subject : str ; required
     - content : str 
    @returns {MailSentResult} : send result

    '''
    # the receiver's address
    if not payload.get('addressee'):
      return {
        'code':501,
        'message':'The addressee address is empty!'
      }
    
    if not payload.get('subject'):
      return {
        'code':502,
        'message':'The mail subject is empty!'
      }

    if not payload.get('content'):
      payload['content'] = payload['subject']
    
    try:
      message = self.__get_message(payload)
      self.connection.sendmail(self.config['addresser'],payload['addressee'],message)
      self.connection.quit()
    except Exception as e:
      return {
        'code':503,
        'message':'%s' % e
      }

    return {
      'code':200,
      'message':'success'
    }

  def __get_message(self,payload):

    message = EmailMessage()
    
    message['subject'] = payload['subject']
    # the last string must be from mail address
    from_with_nickname = '%s <%s>' % (self.config['addresser_name'],self.config['addresser']) 
    message['from'] = Header(from_with_nickname)

    if type(payload['addressee'])==str:
      message['to'] = Header(payload.get('addressee_name',payload['addressee']))
    else:
      message['to'] = Header(','.join(payload['addressee']))

    # support html document
    message.set_content(payload['content'],'html','utf-8')
    return message.as_string()

# singleton mode
BluesMailer = __BluesMailer()   



import sys,os,re,json
from seleniumwire import webdriver
from seleniumwire.utils import decode
from .BluesChrome import BluesChrome
from .http.BluesHTTP import BluesHTTP 

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesFiler import BluesFiler  
from util.BluesURL import BluesURL   
from util.BluesConsole import BluesConsole    

class BluesProxyChrome(BluesChrome):
  # https://pypi.org/project/selenium-wire/
  def __init__(self,config={},arguments={},experimental_options={}):
    super().__init__(config,arguments,experimental_options)
    self.http = BluesHTTP(self.driver)

  def after_created(self):
    self.__set_wire()
    super().after_created()

  def __set_wire(self):
    wire = self.config.get('wire')
    if not wire:
      return
    for key,value in wire.items():
      setattr(self.driver,key,value)

  def get_driver_creator(self):
    '''
    @description : get the webdirver isntance
    '''
    return webdriver

  def get_messages(self,config={}):
    '''
    @description : get all request and response messges
    '''
    config['request_only']=False
    return self.http.get_messages(config)
  
  def save_messages(self,config,file=''):
    '''
    @description : save all request and response messges
    '''
    messages = self.get_messages(config)
    return self.http.save_to_json(messages,file)
    
  def get_requests(self,config={}):
    '''
    @description : get all http messages without response' body
    '''
    config['request_only']=True
    return self.http.get_messages(config)

  def save_requests(self,config,file=''):
    requests = self.get_requests(config)
    return self.http.save_to_json(requests,file)

  def get_cookie_requests(self,config={}):
    '''
    @description : get all requests filter by cookie
    '''
    config['request_only']=True
    return self.http.get_cookie_requests(config)

  def save_cookie_requests(self,config,file=''):
    requests = self.get_cookie_requests(config)
    return self.http.save_to_json(requests,file)

  def get_cookie(self,config={}):
    config['max_count']=1
    requests = self.get_cookie_requests(config)
    if not requests:
      return None
    return requests[0]['headers'].get('Cookie')
  
  def save_cookie(self,config,file=''):
    cookie = self.get_cookie(config)
    return self.http.save_to_txt(cookie,file)




from selenium.webdriver.common.keys import Keys
from .BluesElementAction import BluesElementAction

class BluesFormAction():

  SHORTCUTS={
    'select':Keys.CONTROL+"A",
    'copy':(Keys.CONTROL+"A",Keys.CONTROL+"C"),
    'paste':Keys.CONTROL+"V",
    'cut':(Keys.CONTROL+"A",Keys.CONTROL+"X"), # 拼接为同时使用
    'clear':(Keys.CONTROL+"A",Keys.DELETE), # 元组元素为依次使用
    'space':Keys.SPACE,
    'enter':Keys.ENTER,
  }

  def __init__(self,driver):
    self.driver = driver
    self.element_action = BluesElementAction(driver)

  def shortcut(self,selector,shortcut):
    texts = self.SHORTCUTS.get(shortcut)
    if not texts:
      return 
    
    if type(texts) == str:
      texts = (texts,)

    # multi shortcuts must be type one by one 
    for text in texts:
      self.__type(selector,text) 
    
  def __type(self,selector,texts):
    '''
    @description : type text
    @param {str} selector : css selector
    @param {str|tuple|list} texts
    '''
    web_element = self.element_action.wait(selector)
    input_texts = (texts,) if type(texts)==str else texts 
    web_element.send_keys(*input_texts)

  def write(self,selector,texts):
    '''
    @description : type text
    @param {str} selector : css selector
    @param {str|tuple|list} texts
    '''
    self.shortcut(selector,'clear')
    self.__type(selector,texts)

  def write_after(self,selector,texts):
    self.__type(selector,texts)
  
  def write_file(self,selector,texts):
    self.__type(selector,texts)
  
  def empty(self,selector):
    self.shortcut(selector,'clear')


from selenium.webdriver.support.select import Select
from .BluesElementAction import BluesElementAction

# 提供下拉选择相关功能
class BluesSelectAction():
 
  def __init__(self,driver):
    self.driver = driver
    self.element_action = BluesElementAction(driver)
  
  def get_options(self,selector):
    '''
    @description 获取select选项/选中对象list
    @param {str} selector
    @returns {list}
    '''
    locator = self.element_action.wait(selector)
    options = Select(locator).options
    if options:
      return self.__get_option_items(options)
    else:
      return None
      
  def get_selected_options(self,selector):
    locator = self.element_action.wait(selector)
    options = Select(locator).all_selected_options
    if options:
      return self.__get_option_items(options)
    else:
      return None

  def get_frist_selected_option(self,selector):
    locator = self.element_action.wait(selector)
    option = Select(locator).first_selected_option # return WebElement
    if option:
      return self.__get_option_items([option])[0]
    else:
      return None

  def __get_option_items(self,option_elements):
    items = []
    for option in option_elements:
      items.append({
        'label':option.get_attribute('label'),
        'value':option.text,
      })
    return items

  def set_by_index(self,selector,index=0):
    '''
    @descrition : set a value as the selected value
    @param {int} index 
    '''
    locator = self.element_action.wait(selector)
    return Select(locator).select_by_index(index)
  
  def set_by_value(self,selector,value=''):
    locator = self.element_action.wait(selector)
    return Select(locator).select_by_value(value)

  def set_by_text(self,selector,text=''):
    locator = self.element_action.wait(selector)
    return Select(locator).select_by_visible_text(text)




from .parts.BluesElementAction import BluesElementAction
from .parts.BluesFormAction import BluesFormAction  
from .parts.BluesWindowAction import BluesWindowAction   
from .parts.BluesDocumentAction import BluesDocumentAction    
from .parts.BluesCookieAction import BluesCookieAction     
from .parts.BluesEventAction import BluesEventAction      
from .parts.BluesMoverAction import BluesMoverAction       
from .parts.BluesFrameAction import BluesFrameAction        
from .parts.BluesJavaScriptAction import BluesJavaScriptAction         
from .parts.BluesSelectAction import BluesSelectAction          
from .parts.BluesAlertAction import BluesAlertAction           

class BluesDriverAction():

  def __init__(self,driver):
    self.javascript = BluesJavaScriptAction(driver)
    self.element = BluesElementAction(driver)
    self.form = BluesFormAction(driver)
    self.window = BluesWindowAction(driver)
    self.document = BluesDocumentAction(driver)
    self.cookie = BluesCookieAction(driver)
    self.event = BluesEventAction(driver)
    self.mover = BluesMoverAction(driver)
    self.frame = BluesFrameAction(driver)
    self.select = BluesSelectAction(driver)
    self.alert = BluesAlertAction(driver)


from .BluesElementAction import BluesElementAction

# 提供frame相关功能
class BluesFrameAction():
 
  def __init__(self,driver):
    self.driver = driver
    self.element_action = BluesElementAction(driver)

  '''
  @description frame 窗口切换
  @param {str} frame_selector : iframe selector
  '''
  def switch(self,frame_selector):
    locator = self.element_action.wait(frame_selector)
    self.driver.switch_to.frame(locator)
      
  def quit(self):
    '''
    @description : back to the default partent window
    '''
    self.driver.switch_to.default_content()

  def wrap(self,frame_selector,callback):
    '''
    @description : switch to iframe, execute the callback, then back to main window
    @param {str} frame_selector : iframe's css selector
    @param {function} callback : ececute in the iframe window
    '''
    self.switch(frame_selector)
    callback()
    self.quit()
      
  






import time
# Provides methods for handling three pop-ups
class BluesAlertAction():
 
  def __init__(self,driver):
    self.driver = driver

  '''
  @description frame 切换到alert，不区分三种弹框类型
  @returns {WebElement} alert对象
  '''
  def switch(self):
    return self.driver.switch_to.alert

  '''
  @description 接受-并关闭弹框,the driver will back to main window automatically
  @param {string} text prompt框输入文本
  '''
  def accept(self,text=None):
    alert = self.switch()
    if text!=None:
      alert.send_keys(text)
    alert.accept()
      
  '''
  @description 取消-并关闭弹框
  @param {string} text prompt框输入文本
  '''
  def dismiss(self,text=None):
    alert = self.switch()
    if text!=None:
      alert.send_keys(text)
    alert.dismiss()






import time
from .BluesConsole import BluesConsole

class BluesDateTime():

  spend = 0

  @classmethod
  def count_down(cls,payload):
    '''
    @description : count down
    @param {int} payload.duration  : duration seconds
    @param {int} payload.interval  : interval seconds
    @param {str} payload.title  : title
    @param {bool} payload.printable  : print in the console
    '''

    duration = payload.get('duration',10)
    interval = payload.get('interval',1)
    title = payload.get('title','coutdown')
    printable = payload.get('printable',True)
    
    if printable: 
      BluesConsole.wait('%s : %s' % (title,duration-cls.spend))

    time.sleep(interval) 
    cls.spend+=interval
    if cls.spend < duration:
      cls.count_down(payload)
    else:
      cls.spend=0


import pymysql,traceback

# using the singleton to make sure using the only one cursor
class BluesMySQLExecutor():

  def __init__(self,account):
    '''
    @description : connect and execute sql
    @param {dict} account : MySQL account
    '''
    self.connector = self.__get_connector(account)
    self.cursor = self.connector.cursor() 
  
  def get(self,sql):
    return self.__fetchall(sql)

  def post(self,sql,values=None):
    return self.__execute(sql,values)

  def put(self,sql,values=None):
    return self.__execute(sql,values)
  
  def delete(self,sql):
    return self.__execute(sql)

  def __get_connector(self,account):
    return  pymysql.connect(
      host=account['host'],
      user=account['user'],
      password=account['password'],
      database=account['database'],
      cursorclass=pymysql.cursors.DictCursor,  # 返回数组类型数据
      )

  def __execute(self,sql,values=None):
    '''
    @description : insert/update/delete 
    @prams {str} sql : sql statement (with or without template)
    @params {tuple[]} values : multi real values
    @demo use placeholder
      - sql="insert into ics_test (name,age) values (%s,%s)"
      - execute: cursor.execute(sql,[('blues',18),('liu',12)])
    '''
    try:
      invoker_info = traceback.extract_stack()
      invoker = invoker_info[-2][2]
      
      # use executemany only when values is two-dimensional array
      if values and self.__is_series(values[0]):
        count=self.cursor.executemany(sql,values)
      else:
        count=self.cursor.execute(sql,values)

      self.connector.commit()

      result = {
        'code':200,
        'count':count,
        'sql': sql,
      }
      if invoker=='post':
        # if insert many rows, the row_id is the first inserted row's id (not the last inserted one)
        result['row_id'] = self.cursor.lastrowid

      return result

    except Exception as e:
      return {
        'code':500,
        'exception':e,
        'sql': sql,
      }

  def __is_series(self,value):
    return isinstance(value, (list, tuple))

  def __fetchall(self,sql):
    '''
    @description Query rows of data
    @param {str} sql : Complete sql statement
    @returns {SQLResult} 
    '''
    try:
      self.cursor.execute(sql)
      rows=self.cursor.fetchall()
      # 立即提交，否则轮询会有缓存
      self.connector.commit()
      return {
        'code':200,
        'count':len(rows),
        'data':rows if rows else None, # 无数据返回空元组转为None
        'sql': sql,
      }
    except Exception as e:
      return {
        'code':500,
        'exception':e,
        'sql': sql,
      }








import sys,os,re,time
from .BluesChrome import BluesChrome

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesFiler import BluesFiler  
from util.BluesURL import BluesURL   
from config.BluesConfig import BluesConfig    

class BluesCookieChrome(BluesChrome):

  def __init__(self,config={},arguments={},experimental_options={}):
    super().__init__(config,arguments,experimental_options)

  '''
  @description : support to login by default cookie
  '''
  def get_with_cookie(self,url,cookies=''):
    '''
    @description : get a page afater add cookies
    @param {dict|str} cookies
    '''
    if cookies:
      self.action.cookie.add_cookies(cookies) 
    self.driver.get(url)

  def get_with_cookie_file(self,url,file=''):
    default_file = BluesConfig.get_download_http_domain_file(self.driver.current_url,'txt')
    file_path = file if file else default_file
    if file_path:
      cookies = BluesFiler.read(file_path)
    else:
      cookies = ''
    self.get_with_cookie(url,cookies)



import sys,os,re,subprocess,time
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.options import Options as ChromeOptions

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesPowerShell import BluesPowerShell 
from sele.browser.BluesChrome import BluesChrome  

# debug模式下chrome被PS提前打开，不是由webdriver启动，wire无法实现代理
class BluesDebugChrome(BluesChrome):
  # cover the parent options.detach and prefs can't be set in debug mode
  EXPERIMENTAL_OPTIONS = {}

  def __init__(self,config={},arguments={},experimental_options={}):
    '''
    @description : Get a Chrome Driver instance
    @param {dict} config
    @param {list} arguments
    @param {dict} experimental_options
    '''
    self.chrome_exe = BluesPowerShell.get_env_value('CHROME_EXE')
    self.process_args=''
    self.__validate(arguments)
    super().__init__(config,arguments,experimental_options)

  def before_created(self,config,arguments,experimental_options):
    '''
    @description cover the parent method
    '''
    super().before_created(config,arguments,experimental_options)
    self.__set_args_options()
    self.__restart()

  def __validate(self,arguments):
    error = self.__get_env_error()
    if error:
      raise Exception(error)

    error = self.__get_args_error(arguments)
    if error:
      raise Exception(error)

  def quit(self):
    '''
    @description : close the driver ane the chrome, the driver.quit can't close the chrome that opened by ps script
      don't dirver.quit() or the user login session will be removed
    '''
    self.__close_chrome()

  def __get_env_error(self):
    if not self.chrome_exe:
      return 'Env vairable chrome_exe missing'
    else:
      return ''

  def __get_args_error(self,arguments):
    '''
    @description : Determines whether the entry is legal
    '''
    if not arguments.get('--user-data-dir'):
      return 'Parameter arguments.--user-data-dir missing'
      
    if not arguments.get('--remote-debugging-address'):
      return 'Parameter arguments.--remote-debugging-address missing'

    return ''

  def __set_args_options(self):
    args = ''
    for key,value in self.arguments.items():
      arg_value = '%s=%s' % (key,value) if value else key
      if key == '--remote-debugging-address':
        # must set port or the driver can't connect to the browser
        args+='--remote-debugging-port=%s' % value.split(':')[1]
      else:
        args+=' %s ' % arg_value
    
    self.process_args = args

    # cover the all default options (these default options can't be used to connect to a existed debuug chrome)
    self.experimental_options['debuggerAddress'] = self.arguments['--remote-debugging-address']

  def __close_chrome(self):
    BluesPowerShell.stop_process('chrome')

  def __restart(self):
    '''
    @description : repopen the specific chrome (stop all chrome processes)
    '''
    self.__close_chrome()
    result = BluesPowerShell.start_process(self.chrome_exe,self.process_args)
    if result['code']!=200:
      raise Exception('Chrome restart failure %s' % result['message'])

  



from selenium.webdriver import ActionChains
from .BluesElementAction import BluesElementAction

# 提供元素拖拽功能
class BluesMoverAction():
  def __init__(self,driver):
    self.driver = driver
    self.element_action = BluesElementAction(driver)
    self.chains = ActionChains(driver)

  '''
  @description 按坐标(左上角)移动
  @param {str} selector driver元素
  @params {tuple} position (x,y) 左上角的坐标
  '''
  def position(self,selector,position):
    locator = self.element_action.wait(selector)
    args = (locator ,position[0],position[1])
    self.chains.drag_and_drop_by_offset(*args).perform()

  '''
  @description 按指定偏移量移动
  @param {str} selector driver元素
  @params {tuple} offset (offset_x,offset_y) 移动的尺寸（注意不是坐标），负数是向左/上移动
  '''
  def offset(self,selector,offset):
    coordinate = self.get_coordinate(selector)
    position = (coordinate[0]+offset[0],coordinate[1]+offset[1])
    self.position(selector,position)

  '''
  @description 滑块移动到容器右侧
  @param {str} target : the target container element's css selector
  @param {str} slider : css selector
  '''
  def slide(self,target,slider,direction='right'):
    # 获取容器右侧坐标
    target_coordinate = self.get_coordinate(target)
    slider_coordinate = self.get_coordinate(slider)
    x = slider_coordinate['left'] # init x-axis position
    y = slider_coordinate['top'] # init y-axis position

    if direction == 'left' or direction == 'right':
      x = target_coordinate[direction] 
    elif direction == 'top' or direction == 'bottom':
      y = target_coordinate[direction] 

    self.position(slider,(x,y))

  def center(self,target,slider):
    '''
    @description 滑块移动到容器内(默认到容器中间)
    @param {str} target : the target container element's css selector
    @param {str} slider : css selector
    '''
    target_locator = self.element_action.wait(target)
    slider_locator = self.element_action.wait(slider)
    # the first arg is the slided element
    self.chains.drag_and_drop(slider_locator,target_locator).perform()

  def get_coordinate(self,selector):
    '''
    @description : get element's position and size, the positon is base on window
    @param {str} selector : css selector
    @returns {dict}
    '''
    coordinate = {}
    locator = self.element_action.wait(selector)
    coordinate['left'] = locator.location.get('x')
    coordinate['top'] = locator.location.get('y')
    coordinate['width'] = locator.size.get('width')
    coordinate['height'] = locator.size.get('height')
    coordinate['right'] = coordinate['left']+coordinate['width']
    coordinate['bottom'] = coordinate['top']+coordinate['height']
    return coordinate





from setuptools import setup,find_packages

setup(
  name="blues_lib", # package name
  version="1.1.17", # package version
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  packages=find_packages(), # package module
  # add from requirement.txt,本地测试注释所有包，不能从镜像立即安装
  install_requires=[
    #'Pillow>=10.3.0',
    #'Requests>=2.31.0',
    #'selenium>=4.20.0',
    #'selenium_wire>=5.1.0',
    #'setuptools>=47.1.0',
    #'webdriver_manager>=4.0.0',
  ] # package dependency
)


import sys,os,re,time
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions 

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesConsole import BluesConsole  

# 提供元素选择功能
class BluesElementAction():

  def __init__(self,driver):
    self.driver = driver

  # === part 1:  find element === #
  def wait(self,selector,timeout=10,wait_type='presence'):
    '''
    @description : wait and return a WebElement
    @param {str}  selector : css selector
    @param {int} timeout : Maximum waiting time (s)
    @param {str} wait_type
    @returns {bool} if the element exists in the maximum time
    '''
    wait_func = self.__get_wait_func(selector,wait_type)
    try:
      BluesConsole.wait(selector,'Wait for element')
      return WebDriverWait(self.driver,timeout=timeout).until(wait_func)
    except Exception as e:
      return None

  def wait_removed(self,selector,timeout=20,step=1):
    '''
    @description : wait a element to be removed
    @param {str}  selector : css selector
    @param {int} timeout : Maximum waiting time (s)
    @param {int} set : wait seconds every time
    @returns {bool} if the element removed in maximum time
    '''
    for i in range(timeout):
      time.sleep(step)
      if not self.exists(selector):
        return True 
    return False

  def get(self,selector):
    '''
    @description : get the first WebElement, if none exists, an exception is thrown - dont't wait 1 second
    @param {str} selector : css selector
    @returns {WebElement}
    '''
    return self.driver.find_element(By.CSS_SELECTOR,selector)

  def get_all(self,selector):
    '''
    @description : get the first WebElement, if none exists, return empty list - dont't wait 1 second
    @param {str} selector : css selector
    @returns {WebElement[] | []}
    '''
    return self.driver.find_elements(By.CSS_SELECTOR,selector)

  # === part 2:  get element info === #
  def get_attr(self,selector,key):
    '''
    @description : get element's attribute
    @param {str} selector : css selector
    @param {str} key : attribute name, all HTML DOM attributes can be accessed
      - innerHTML
      - innerText
    @returns {str}
    '''
    web_element = self.wait(selector)
    return web_element.get_attribute(key)

  def get_text(self,selector):
    web_element = self.wait(selector)
    return web_element.text

  # === part 3:  check element status === #
  def exists(self,selector):
    '''
    @description : Determines whether the element exists
    @param {str} selector : css selector
    @returns {bool}
    '''
    eles = self.get_all(selector)
    return True if eles else False

  def is_presence(self,selector,timeout=1):
    return self.__get_status(selector,timeout,'presence')

  def is_visibility(self,selector,timeout=1):
    return self.__get_status(selector,timeout,'visibility')

  def is_clickable(self,selector,timeout=1):
    return self.__get_status(selector,timeout,'clickable')
  
  def is_enabled(self,selector):
    return self.get(selector).is_enabled()

  def is_displayed(self,selector):
    return self.get(selector).is_displayed()

  def is_selected(self,selector):
    return self.get(selector).is_selected()

  # === part last:  tool func === #
  def __get_status(self,selector,timeout,wait_type):
    '''
    @description : Determines whether the element exists
    @param {str} selector : css selector
    @param {str} wait_type
    @returns {bool}
    '''
    try:
      self.wait(selector,timeout,wait_type)
      return True
    except Exception as e:
      return False

  def __get_wait_func(self,selector,wait_type='presence'):
    '''
    @description 获取条件判断函数
    '''
    # 内置条件函数入参tuple格式
    css_selector = (By.CSS_SELECTOR,selector)

    funcs = {
      'presence':expected_conditions.presence_of_element_located, # 存在于DOM中（可见或不可见）
      'visibility':expected_conditions.visibility_of_element_located, # 存在于DOM中且可见（非隐藏状态，不一定再当前窗口中）
      'clickable':expected_conditions.element_to_be_clickable, # 存在于DOM中且可见，可点击（在当前窗口中）
      'selected':expected_conditions.element_to_be_selected, # 专用于下拉选择项，存在与DOM中，且元素是被选中
    }

    func = funcs.get(wait_type,funcs['presence']) 
    return func(css_selector)

    
    




import requests
# 提供js相关功能
class BluesJavaScriptAction():
  
  PLUGINS = {
    '$':'https://libs.baidu.com/jquery/2.0.0/jquery.min.js',
    #'_':'https://cdn.bootcdn.net/ajax/libs/lodash.js/4.17.21/lodash.core.js',
  }

  def __init__(self,driver):
    self.driver = driver

  def load_jquery(self):
    self.load_plugins({
      '$':self.PLUGINS['$']
    })

  def load_plugins(self,plugins):
    '''
    @description : load relay plugin auto
    @param {dict} plugins 
      - key : the global var name
      - value : the plugin's cdn
    '''
    for plugin_var,plugin_cdn in plugins.items():
      if not self.is_var_exists(plugin_var):
        self.load(plugin_cdn )

  def execute(self,script):
    '''
    @description 执行js代码
    @param {str} script js脚本
    @returns {str} the js script's return value
    '''
    return self.driver.execute_script(script)
  
  def execute_async_script(self,script):
    '''
    @description async 执行js代码
    @param {str} script js脚本
    @returns {str} the js script's return value
    '''
    return self.driver.execute_async_script(script)

  def execute_external_script(self,url):
    '''
    @description : run the external script 
    @param {str} url : the script file path
    @returns {str} : the js return value
    '''
    response = requests.get(url = url)
    return self.execute(response.text)
  

  def html(self,selector,html=None):
    '''
    @description : get or set html
    @param {str} selector css选择器
    @param {str} html 写入文本或HTML
    @returns {str|None} return dom html when as getter
    '''
    if html==None:
      script = """
       return $('{}').html()
      """.format(selector)
    else:
      script = """
        $('{}').html('{}')
      """.format(selector,html)
    return self.execute(script)

  def text(self,selector,text=None):
    '''
    @description : get or set text
    @param {str} selector css选择器
    @param {str} text 
    @returns {str|None} return dom html when as getter
    '''
    if text==None:
      script = """
       return $('{}').text()
      """.format(selector)
    else:
      script = """
        $('{}').text('{}')
      """.format(selector,text)
    return self.execute(script)


  def value(self,selector,value=None):
    '''
    @description : set or get form controller's value
    '''
    if value==None:
      script = """
        return $('{}').val()
      """.format(selector)
    else:
      script = """
        $('{}').val('{}')
      """.format(selector,value)
    return self.execute(script)

  def html_after(self,selector,html):
    '''
    @description PM04 追加元素内容
    @param {str} selector css选择器
    @param {str} html 写入文本或HTML
    '''
    script = """
      $('{}').append('{}')
    """.format(selector,html)
    return self.execute(script)

  def text_after(self,selector,text):
    '''
    @description : append text 
    @param {str} selector css选择器
    @param {str} text 
    '''
    script = """
      $('{}').text($('{}').text()+'{}')
    """.format(selector,selector,text)
    return self.execute(script)


  def value_after(self,selector,value):
    '''
    @description : set form controller's value
    '''
    script = """
      $('{}').val($('{}').val()+'{}')
    """.format(selector,selector,value)
    self.execute(script)

  def attr(self,selector,key_or_items):
    '''
    @description : get or set element's attribute
    @param {str} selector : element's css selector
    @param {str|dict} key_or_items : 
      - str : getter , such as 'name'
      - dict : setter , such as {'data-a','1','data-b','2'}
    @returns {str|None} : return the attr's value in gtter
    '''
    if type(key_or_items) == str:
      # must use return or can't get the js return value
      script = "return $('%s').attr('%s')" % (selector,key_or_items)
    else:
      attrs=''
      for key,value in key_or_items.items():
        attrs+=".attr('%s','%s')" % (key,value)
      script = "$('%s')%s" % (selector,attrs)
    return self.execute(script)
   
  def css(self,selector,key_or_items ):
    '''
    @description : get or set element's style
    @param {str} selector : element's css selector
    @param {str|dict} key_or_items : 
      - str : getter , such as 'name'
      - dict : setter , such as {'data-a','1','data-b','2'}
    @returns {str|None} : return the css style's value in gtter
    '''
    if type(key_or_items) == str:
      script = "return $('%s').css('%s')" % (selector,key_or_items)
    else:
      styles=''
      for key,value in key_or_items.items():
        styles+=".css('%s','%s')" % (key,value)
      script = "$('%s')%s" % (selector,styles)
    return self.execute(script)

  def click(self,selector):
    script = "$('%s').click()" % selector
    self.execute(script)


  def scroll_top(self,top=0,scroll_selector='document'):
    self.__scroll(top,scroll_selector,'y')
  
  def scroll_left(self,top=0,scroll_selector='document'):
    self.__scroll(top,scroll_selector,'x')

  def __scroll(self,top=0,scroll_selector='document',direction='y'):
    '''
    @description : scroll in y-axis
    @param {int} top : the offset to top
      0 - to top
      -1 - to bottom
    @param {str} scroll_selector : the scroll element
    '''
    methods = {
      'x':'scrollLeft',
      'y':'scrollTop',
    }
    outer_selector = self.__get_scroll_element(scroll_selector)
    if top == -1:
      offset = '$(%s).height()' % outer_selector
    else:
      offset = top
    
    method = methods.get(direction,methods['y'])
    script = "$(%s).%s(%s);" % (outer_selector,method,offset)
    return self.execute(script)

  def scroll_element_to_window(self,content_selector,scroll_selector='document'):
    '''
    @description : scroll the content element into the window 
    @param {str} content_selector : the element will be shown in window
    @param {str} scroll_selector : the scroll element
    '''
    outer_selector = self.__get_scroll_element(scroll_selector)
    script = "$(%s).scrollTop($('%s').offset().top-100);" % (outer_selector,content_selector)
    self.execute(script)

  def __get_scroll_element(self,selector):
    return selector if selector == 'document' else "'%s'" % selector

  def remove(self,selector):
    '''
    @description 移除元素
    @param {str} selector css选择器
    '''
    script = """
      $('{}').remove()
    """.format(selector)
    return self.execute(script)

  def empty(self,selector):
    '''
    @description 清空元素内容
    @param {str} selector css选择器
    '''
    script = """
      $('{}').empty()
    """.format(selector)
    return self.execute(script)

  def load(self,url):
    '''
    @description : 导入线上/本地 js脚本
    @param {str} url : the scirpt file's path
    '''
    mark_text = "selenium-dynamic-script:%s" % url
    create_script = "var sele_script=document.createElement('script');sele_script.src='%s';" % url
    onload_script = "sele_script.onload=function(){console.log('%s loaded by the selenium!')};" % url
    append_script = "var head=document.getElementsByTagName('head')[0];head.appendChild(sele_script);"
    load_script = "%s %s %s" % (create_script,onload_script,append_script)
    # wait the script loaded, Determine whether it is complete according to the dynamic insertion mark 'selenium-dynamic-script'
    return self.execute(load_script)

  def is_var_exists(self,var):
    '''
    @description : determine weather global vars exists
    @param {str|list|tuple} var : consider some var has alias
    @returns {bool}
    '''
    vars = [var] if type(var)==str else var 
    condition = ''
    for key in vars:
      condition+='|| window["%s"]' % key
    condition = condition[2:]
    script = "var global_var = %s; return !!global_var;" % condition
    return self.execute(script)

  def open(self,url,name=''):
    '''
    @description : open a url in a new tab, and swith the driver to the new tab
    '''
    script = "window.open('%s','%s')" % (url,name)
    self.execute(script)



from PIL import Image
from .BluesFiler import BluesFiler

class BluesImager:

  @classmethod
  def download(cls,urls,directory,target_size=None):
    success = lambda local_image : cls.resize(local_image,target_size)
    return BluesFiler.download(urls,directory,success)

  @classmethod
  def __get_scale_size(cls,current_size,target_size):
    '''
    @description : Gets the smallest scale size
    @param {tuple} current_size : current the image's real size
    @param {tuple} target_size : the target size ,the will may be not scale 
    @returns {tuple} new size
    '''
    (current_width,current_height) = current_size
    (target_width,target_height) = target_size

    if not target_size or (not target_width and not target_height):
      return current_size

    if target_width and not target_height:
      target_height = int(target_width*current_height/current_width) 

    if target_height and not target_width:
      target_width = int(target_height*current_width/current_height) 

    # get the scale min target size
    if target_width and target_height:
      width_ratio = target_width/current_width
      height_ratio = target_height/current_height
      # With a bigger scale as the standard
      if width_ratio>height_ratio:
        target_height = int(target_width*current_height/current_width) 
      else:
        target_width = int(target_height*current_width/current_height) 
    return (target_width,target_height)

  @classmethod
  def resize(cls,local_image,target_size):
    '''
    @description : set image's size
    @param {str} local_image 
    @param {tuple} target_size : (width,height)
    '''

    # 等比例设置
    with Image.open(local_image) as im:

      current_size = im.size
      scale_size = cls.__get_scale_size(current_size,target_size)

      copy_image = cls.__copy(im,local_image,'','original')
      new_image = im.resize(scale_size)

      # convert rgba to rgb; or can't save to jpeg
      rgb_image = new_image.convert('RGB')
      rgb_image.save(local_image)

      return {
        'original_image':copy_image,
        'original_size':current_size,
        'target_size':target_size,
        'size':scale_size
      }

  @classmethod
  def copy(cls,local_image,new_name='',prefix='',suffix='',separator='-'):
    with Image.open(local_image) as im:
      return cls.__copy(im,local_image,new_name,prefix,suffix,separator)

  @classmethod
  def __copy(cls,im,local_image,new_name='',prefix='',suffix='',separator='-'):
    img_copy = im.copy()
    path = BluesFiler.get_rename_file(local_image,new_name,prefix,suffix,separator)
    img_copy.save(path)
    return path





import os,requests,json
from datetime import datetime,timedelta

class BluesFiler:

  @classmethod
  def removedirs(cls,directory):
    '''
    @description Remove all child dir and files
    @param {string} directory 
    '''
    removed_count = 0
    for root, dirs, files in os.walk(directory):
      for file in files:
        os.remove(os.path.join(root, file))
        removed_count +=1
      for dir in dirs:
        os.rmdir(os.path.join(root, dir))
        removed_count +=1
    return removed_count

  @classmethod
  def download(cls,urls,directory,success=None,error=None):
    '''
    @description Download multi files
    @param {list} urls files' remote url
    @param {string} directory : Local directory to save the downloaded files
    @param {function} success : Callback function called on success
    @param {function} error : Callback function called on failure
    @returns {dict} complex result
    '''

    result = cls.__get_result()

    if not urls:
      return result 

    for url in urls:
      # download the image
      (code,file_or_msg) = cls.download_one(url,directory)
      if code == 200:
        item = {
          'url':url,
          'file':file_or_msg,
          'callback_value':None
        }
        if success:
          item['callback_value'] = success(file_or_msg)
        result['success']['count']+=1
        result['success']['files'].append(item)
        result['files'].append(file_or_msg)
        result['code'] = 200
      else:
        item = {
          'url':url,
          'message':file_or_msg,
          'callback_value':None
        }
        if error:
          item['callback_value'] = error(str(e))
        result['error']['count']+=1
        result['error']['files'].append(item)
    
    return result 

  @classmethod
  def download_one(cls,url,directory):
    '''
    @description : download one file
    @param {str} url : file's remote url
    @param {str} directory : The dir to save the download file
    '''
    try:
      # Ensure directory existence
      cls.makedirs(directory)
      # Keep the file name unchanged
      file_name = url.split('/')[-1]
      local_file = directory+'/'+file_name

      # 永远保持覆盖
      res=requests.get(url)
      res.raise_for_status()
      with open(local_file,'wb') as f:
        f.write(res.content)
        f.close()
        return (200,local_file) 

    except Exception as e:
      return (500,str(e))

  @classmethod
  def read(cls,file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
      return file.read()

  @classmethod
  def read_json(cls,file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
      data = json.load(file)
      return data

  @classmethod
  def write_json(cls,file,data,indent=2):
    return BluesFiler.write(file,json.dumps(data,indent=indent))

  @classmethod
  def __get_result(cls):
    return {
      'code':500,
      'files':[],
      'success':{
        'count':0,
        'files':[],
      },
      'error':{
        'count':0,
        'files':[],
      },
    }

  @classmethod
  def write(cls,file_path,text,mode='w'):
    '''
    @description : write text to file
    @param {str} file_path : file's path
    @param {str} text : content
    @param {str} mode : write mode
      - 'w' : clear the history content
      - 'a' : append text
    @returns {None} 
    '''
    dir_path = os.path.dirname(file_path)
    if dir_path!='.':
      cls.makedirs(dir_path)

    with open(file_path,mode,encoding='utf-8') as file:
      file.write(text)
      return file_path

  @classmethod
  def exists(cls,path):
    '''
    @description : Does a dir or file exist
    @param {str} path
    @returns {bool} 
    '''
    return os.path.exists(path)

  @classmethod
  def makedirs(cls,path):
    '''
    @description : Create dirs (support multilevel directory) if they don't exist
    @param {str} path : multilevel dir
    @returns {None}
    '''
    if not cls.exists(path):
      os.makedirs(path)

  @classmethod
  def get_rename_file(cls,file_path,new_name='',prefix='',suffix='',separator='-'):
    '''
    @description : get the new file name path
    '''
    path_slices = file_path.split('/')
    original_name = path_slices[-1]
    copy_name = new_name if new_name else original_name
    if prefix:
      copy_name = prefix+separator+copy_name
    if suffix:
      copy_name = copy_name+separator+suffix
    path_slices[-1]=copy_name
    copy_path='/'.join(path_slices)
    return copy_path

  @classmethod
  def removefiles(cls,directory,retention_days=7):
    '''
    @description : clear files before n days
    @param {str} directory
    @param {int} retention_days : default 7
    @returns {int} deleted files count
    '''
    # 转换天数到时间间隔
    threshold = datetime.now() - timedelta(days=retention_days)
    removed_count = 0
    # 遍历目录
    for item in os.scandir(directory):
      try:
        # 获取文件的最后修改时间
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(item.path))
        # 如果文件的最后修改时间早于阈值，则删除文件
        if os.path.isfile(item.path) and file_modified_time < threshold:
          os.remove(item.path)
          removed_count +=1
      except OSError as e:
        pass

    return removed_count




from selenium.webdriver import ActionChains
from .BluesElementAction import BluesElementAction

class BluesEventAction():
  def __init__(self,driver):
    self.driver = driver
    self.element_action = BluesElementAction(driver)
    self.chains = ActionChains(driver)

  def click(self,selector):
    '''
    @description 单击
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.click(locator).perform()
  
  def hold(self,selector):
    '''
    @description 点击按住不放
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.click_and_hold(locator).perform()

  def release(self,selector):
    '''
    @description 释放按下的鼠标
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.release(locator).perform()

  def right_click(self,selector):
    '''
    @description 右击
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.context_click(locator).perform()

  def double_click(self,selector):
    '''
    @description 双击
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.double_click(locator).perform()

  def hover(self,selector):
    '''
    @description 鼠标移入元素，悬浮上方
    @param {str} selector
    '''
    locator = self.element_action.wait(selector)
    self.chains.move_to_element(locator).perform()

  


import sys,os,re,time
from .BluesJavaScriptAction import BluesJavaScriptAction

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesType import BluesType 

# 提供窗口相关功能
class BluesWindowAction():
 
  def __init__(self,driver):
    self.driver = driver
    self.js_action = BluesJavaScriptAction(driver)
    self.history_handles = [self.get_handle()]
  
  def shot(self,file=None):
    '''
    @description : shot screen as a file
    @param {str} file : File save address
    @returns {None} 
    '''
    file_path = file if file else self.__get_default_file()
    shot_status = self.driver.save_screenshot(file_path)
    return file_path if shot_status else ''

  def shot_base64(self):
    '''
    @description : shot screen as base64 string
    @returns {str} : base64 string
    '''
    return self.driver.get_screenshot_as_base64()

  def __get_default_file(self,prefix='screenshot'):
    timestamp = int(time.time()) 
    return './%s-%s.png' % (prefix,timestamp)

  def get_handle(self):
    '''
    @description : get current active tab's handle id
    @returns {str} handle id, such as : 3D31BF6D96E5671253E70BCF33DC7F39
    '''
    return self.driver.current_window_handle

  # 所有窗口句柄 list
  def get_handles(self):
    '''
    @description : get all tab's handle id
    @returns {list} handle id, such as : ['3D31BF6D96E5671253E70BCF33DC7F39']
    '''
    return self.driver.window_handles

  def switch(self,handle):
    '''
    @description window窗口切换（切换浏览器tab页签）
    @param {string} handle 窗口句柄
    '''
    self.driver.switch_to.window(handle)

  def open_tab(self,url,name=''):
    self.js_action.open(url,name)
    new_handle = self.get_handles()[-1]
    self.history_handles.append(new_handle)
    self.switch(new_handle)

  def back_tab(self):
    if len(self.history_handles) > 1:
      prev_handle = self.history_handles[-2]
      self.switch(prev_handle)

  def forward_tab(self):
    current_handle = self.get_handle()
    current_handle_index = BluesType.last_index(self.history_handles,current_handle)
    if len(self.history_handles)-1>current_handle_index:
      next_handle = self.history_handles[current_handle_index+1]
      self.switch(next_handle)

  def get_size(self):
    '''
    @descripton : get current window's size
    @returns {dict} : {'width': 1755, 'height': 946}
    '''
    return self.driver.get_window_size()

  def resize(self,size):
    '''
    @description 修改窗口尺寸
    @param {tuple|list} size (width,height)
    '''
    self.driver.set_window_size(*size)

  def maximize(self):
    '''
    @description : maximize
    '''
    self.driver.maximize_window()

  def minimize(self):
    '''
    @description : hide the chrome window
    '''
    self.driver.minimize_window()

  def fullscreen(self):
    '''
    @description : maximize and hide the tabs
    '''
    self.driver.fullscreen_window()

  def position(self,position):
    '''
    @description 修改窗口位置
    @param {tuple} position (x,y)
    '''
    self.driver.set_window_position(*position)
     
  def refresh(self):
    self.driver.refresh()
      
  def back(self):
    self.driver.back()
      
  def forward(self):
    self.driver.forward()



import sys,os,re,json

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesFiler import BluesFiler  
from util.BluesURL import BluesURL   
from util.BluesConsole import BluesConsole
from config.BluesConfig import BluesConfig 

class BluesHTTP():
  def __init__(self,driver):
    self.driver = driver

  def get_messages(self,config={}):
    '''
    @description : get all http request and response message
    @param {dict} config
     - {int} max_count : the maximum number of request messages, -1 means infinity
     - {str} url_pattern : Only data that matches the url pattern is retrieved
     - {str} cookie_pattern : Only data that matches the cookie pattern is retrieved
     - {function} filter(message) : Only data that meets the filtering conditions is obtained
     - {bool} request_only : If true, without response's body
    @returns {dict[] | None}
    '''
    if not self.driver.requests:
      return None

    messages = []
    max_count = config.get('max_count',-1) # -1 indicates that all data is obtained
    url_pattern = config.get('url_pattern')
    message_filter = config.get('message_filter')
    request_only = config.get('request_only',False)

    for request in self.driver.requests:
      if not request.response:
        continue
      
      # filter by url_pattern
      if url_pattern:
        if not re.search(url_pattern,request.url):
          continue

      message = self.__get_message(request,request_only)
      # filter by user-defined func
      if message_filter:
        if not message_filter(message):
          continue
       
      messages.append(message)

      # break by max_count
      if max_count!=-1 and len(messages)>=max_count:
        break

    if messages:
      return messages
    else:
      BluesConsole.warn(('No matching http message was obtained!',config))
      return None

  def __get_message(self,request,request_only=False):
    return {
      'url':request.url, 
      'path':request.path, 
      'querystring':request.querystring, 
      'method':request.method, 
      'headers':dict(request.headers), 
      'params':request.params, 
      'date':request.date.isoformat(), 
      #'cert':request.cert, 
      'body':self.__get_request_body(request.body), 
      'response':self.__get_response(request.response,request_only), 
    }

  def __get_request_body(self,body):
    return str(body, encoding='utf-8')

  def __get_response(self,response,request_only=False):
    res_message = {
      'status_code':response.status_code, 
      'reason':response.reason, 
      'headers':dict(response.headers), 
      'date':response.date.isoformat(), 
    }
    if not request_only:
      encoding = response.headers.get('Content-encoding', 'identity')
      body = decode(response.body, encoding)
      res_message['body'] = body

    return res_message

  def save_to_txt(self,content,file=''):
    default_file = BluesConfig.get_download_http_domain_file(self.driver.current_url,'txt')
    file_path = file if file else default_file
    return BluesFiler.write(file_path,content)

  def save_to_json(self,messages,file=''):
    '''
    @description : save messages to json file
    @param {any} messages
    @param {str} file : the json file's absolute path, If this value is empty:
      - the dir will be c:/blues_lib/download/cookie/
      - the file will be domain.json
    '''
    default_file = BluesConfig.get_download_http_domain_file(self.driver.current_url,'json')
    file_path = file if file else default_file
    try:
      BluesFiler.write_json(file_path,messages)
      return {
        'code':200,
        'message':'success',
        'file':file_path,
      }
    except Exception as e:
      return {
        'code':500,
        'message':'%s' % e,
      }

  def get_cookie_requests(self,config={}):
    '''
    @description : get all requets filter by cookie
    '''
    domain = BluesURL.get_domain(self.driver.current_url)
    def message_filter(message):
      if not re.search(domain,message['url']):
        return False

      cookie = message['headers'].get('Cookie')
      if not cookie:
        return False

      cookie_pattern = config.get('cookie_pattern')
      if not cookie_pattern:
        return True

      if not re.search(cookie_pattern,cookie):
        return False

      return True

    if not config.get('message_filter'):
      config['message_filter'] = message_filter
    
    requests = self.get_messages(config)
    return requests if requests else None


  


import sys,os,re,subprocess,time
from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.chrome.service import Service as ChromeService
from selenium.webdriver.chrome.options import Options as ChromeOptions

sys.path.append(re.sub('blues_lib.*','blues_lib',os.path.realpath(__file__)))
from util.BluesPowerShell import BluesPowerShell 
from sele.action.BluesDriverAction import BluesDriverAction  

class BluesChrome():

  CONFIG = {
    'url':'https://www.baidu.com', # default open web page
  }

  # https://peter.sh/experiments/chromium-command-line-switches/
  ARGUMENTS = {
    '--start-maximized':'',
    '--no-default-browser-check':'',
    '--disable-notifications':'',
    #'--disable-web-security':'', # 禁用此项，会导致debug模式下也无法存储登录信息
    #'--ignore-certificate-errors':'',
    '--disable-infobars':'',
    '--hide-crash-restore-bubble':'', # 是否打开崩溃前的页面
    '--disable-popup-blocking':'',
    '--disable-extensions-api':'', 
    '--disable-application-install-prompt':'',
    '--no-first-run':'', 
    '--disable-first-run-ui':'',
  }

  EXPERIMENTAL_OPTIONS = {
    'detach':True,
    'prefs' : {
      # 0 - Default, 1 - Allow, 2 - Block
      "profile.default_content_setting_values.notification": 2,
      # Removing the webdriver property from the navigator object,so the browser could not distinguish whether for robots
      "Page.addScriptToEvaluateOnNewDocument":{
        "source": """
          Object.defineProperty(navigator, 'webdriver', {
            get: () => undefined
          })
        """
      }
    }
  }

  def __init__(self,config={},arguments={},experimental_options={}):
    '''
    @description : Get a Chrome Driver instance
    @param {dict} config
    '''
    self.before_created(config,arguments,experimental_options)
    self.created()
    self.after_created()

  def before_created(self,config,arguments,experimental_options):
    self.driver = None
    self.action = None
    self.config = {**self.CONFIG,**config}
    self.arguments = {**self.ARGUMENTS,**arguments} 
    # connect to a exist chrome, can't set detach and prefs options
    if experimental_options.get('debuggerAddress'):
      self.experimental_options = experimental_options
    else:
      self.experimental_options = {**self.EXPERIMENTAL_OPTIONS,**experimental_options}
    self.chrome_driver_exe = BluesPowerShell.get_env_value('CHROME_DRIVER_EXE')

  def created(self):
    '''
    @description : create the driver and action instance
    '''
    chrome_service = self.__get_service()
    chrome_options = self.__get_options()
    web_driver = self.get_driver_creator()
    self.driver = web_driver.Chrome( service = chrome_service, options = chrome_options)
    self.action = BluesDriverAction(self.driver)  

  def after_created(self):
    '''
    @description : do some settings after the driver was created
     provide a hook to make subclass to insert private logic
    '''
    url = self.config.get('url')
    if url:
      self.driver.get(url)
  
  def quit(self):
    self.driver.quit()

  def close(self):
    self.driver.close()

  def get_driver_creator(self):
    '''
    @description : get the webdirver isntance, it will be cover by subclass BluesProxyChrome
    '''
    return webdriver

  def __get_options(self):
    chrome_options = ChromeOptions()
    
    for key,value in self.arguments.items():
      arg_value = '%s=%s' % (key,value) if value else key
      chrome_options.add_argument(arg_value)

    for key,value in self.experimental_options.items(): 
      chrome_options.add_experimental_option(key,value)

    return chrome_options

  def __get_service(self):
    executable_path=self.chrome_driver_exe if self.chrome_driver_exe else ChromeDriverManager().install()
    return ChromeService(executable_path) 



